@@ -22,19 +22,26 @@ show_help() {
   echo "-B --bundle               Create an App bundle (macOS only)"
   echo "-P --portable             Create a portable binary package."
   echo "-O --pgo                  Use profile guided optimizations (pgo)."
+  echo "-U --windows-lua-utf      Use the UTF8 patch for Lua."
   echo "                          macOS: disabled when used with --bundle,"
   echo "                          Windows: Implicit being the only option."
+  echo "-r --release              Compile in release mode."
   echo
 }
 
 main() {
   local platform="$(get_platform_name)"
+  local arch="$(get_platform_arch)"
   local build_dir="$(get_default_build_dir)"
+  local build_type="debug"
   local prefix=/
   local force_fallback
   local bundle
   local portable
   local pgo
+  local patch_lua
+
+  local lua_subproject_path
 
   for i in "$@"; do
     case $i in
@@ -76,6 +83,14 @@ main() {
         pgo="-Db_pgo=generate"
         shift
         ;;
+      -U|--windows-lua-utf)
+        patch_lua="true"
+        shift
+        ;;
+      -r|--release)
+        build_type="release"
+        shift
+        ;;
       *)
         # unknown option
         ;;
@@ -92,20 +107,41 @@ main() {
       portable=""
   fi
 
+  if [[ $CROSS_ARCH != "" ]]; then
+    if [[ $platform == "macos" ]]; then
+      macos_version_min=10.11
+      if [[ $CROSS_ARCH == "arm64" ]]; then
+        cross_file="--cross-file resources/macos/macos_arm64.conf"
+        macos_version_min=11.0
+      fi
+      export MACOSX_DEPLOYMENT_TARGET=$macos_version_min
+      export MIN_SUPPORTED_MACOSX_DEPLOYMENT_TARGET=$macos_version_min
+      export CFLAGS=-mmacosx-version-min=$macos_version_min
+      export CXXFLAGS=-mmacosx-version-min=$macos_version_min
+      export LDFLAGS=-mmacosx-version-min=$macos_version_min
+    fi
+  fi
+
   rm -rf "${build_dir}"
 
   CFLAGS=$CFLAGS LDFLAGS=$LDFLAGS meson setup \
-    --buildtype=release \
+    --buildtype=$build_type \
     --prefix "$prefix" \
+    $cross_file \
     $force_fallback \
     $bundle \
     $portable \
     $pgo \
     "${build_dir}"
 
+  lua_subproject_path=$(echo subprojects/lua-*/)
+  if [[ $patch_lua == "true" ]] && [[ ! -z $force_fallback ]] && [[ -d $lua_subproject_path ]]; then
+    patch -d $lua_subproject_path -p1 --forward < resources/windows/001-lua-unicode.diff
+  fi
+
   meson compile -C "${build_dir}"
 
-  if [ ! -z ${pgo+x} ]; then
+  if [[ $pgo != "" ]]; then
     cp -r data "${build_dir}/src"
     "${build_dir}/src/lite-xl"
     meson configure -Db_pgo=use "${build_dir}"
