@@ -1,29 +1,51 @@
 #include <SDL.h>
+#include <string.h>
 #include <stdbool.h>
 #include <ctype.h>
-#include <dirent.h>
-#include <unistd.h>
 #include <errno.h>
+#include <sys/types.h>
 #include <sys/stat.h>
 #include "api.h"
-#include "dirmonitor.h"
-#include "rencache.h"
+#include "../rencache.h"
 #ifdef _WIN32
   #include <direct.h>
   #include <windows.h>
   #include <fileapi.h>
-#elif __linux__
+  #include "../utfconv.h"
+
+  // Windows does not define the S_ISREG and S_ISDIR macros in stat.h, so we do.
+  // We have to define _CRT_INTERNAL_NONSTDC_NAMES 1 before #including sys/stat.h
+  // in order for Microsoft's stat.h to define names like S_IFMT, S_IFREG, and S_IFDIR,
+  // rather than just defining  _S_IFMT, _S_IFREG, and _S_IFDIR as it normally does.
+  #define _CRT_INTERNAL_NONSTDC_NAMES 1
+  #include <sys/types.h>
+  #include <sys/stat.h>
+  #if !defined(S_ISREG) && defined(S_IFMT) && defined(S_IFREG)
+    #define S_ISREG(m) (((m) & S_IFMT) == S_IFREG)
+  #endif
+  #if !defined(S_ISDIR) && defined(S_IFMT) && defined(S_IFDIR)
+    #define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
+  #endif
+#else
+
+#include <dirent.h>
+#include <unistd.h>
+
+#ifdef __linux__
   #include <sys/vfs.h>
 #endif
+#endif
 
 extern SDL_Window *window;
 
 
 static const char* button_name(int button) {
   switch (button) {
-    case 1  : return "left";
-    case 2  : return "middle";
-    case 3  : return "right";
+    case SDL_BUTTON_LEFT   : return "left";
+    case SDL_BUTTON_MIDDLE : return "middle";
+    case SDL_BUTTON_RIGHT  : return "right";
+    case SDL_BUTTON_X1     : return "x";
+    case SDL_BUTTON_X2     : return "y";
     default : return "?";
   }
 }
@@ -43,7 +65,7 @@ struct HitTestInfo {
 };
 typedef struct HitTestInfo HitTestInfo;
 
-static HitTestInfo window_hit_info[1] = {{0, 0}};
+static HitTestInfo window_hit_info[1] = {{0, 0, 0}};
 
 #define RESIZE_FROM_TOP 0
 #define RESIZE_FROM_RIGHT 0
@@ -104,16 +126,55 @@ static const char *get_key_name(const SDL_Event *e, char *buf) {
     !(e->key.keysym.mod & KMOD_NUM)) {
     return numpad[scancode - SDL_SCANCODE_KP_1];
   } else {
-    strcpy(buf, SDL_GetKeyName(e->key.keysym.sym));
+    /* We need to correctly handle non-standard layouts such as dvorak.
+       Therefore, if a Latin letter(code<128) is pressed in the current layout,
+       then we transmit it as it is. But we also need to support shortcuts in
+       other languages, so for non-Latin characters(code>128) we pass the
+       scancode based name that matches the letter in the QWERTY layout.
+
+       In SDL, the codes of all special buttons such as control, shift, arrows
+       and others, are masked with SDLK_SCANCODE_MASK, which moves them outside
+       the unicode range (>0x10FFFF). Users can remap these buttons, so we need
+       to return the correct name, not scancode based. */
+    if ((e->key.keysym.sym < 128) || (e->key.keysym.sym & SDLK_SCANCODE_MASK))
+      strcpy(buf, SDL_GetKeyName(e->key.keysym.sym));
+    else
+      strcpy(buf, SDL_GetScancodeName(scancode));
     str_tolower(buf);
     return buf;
   }
 }
 
+#ifdef _WIN32
+static char *win32_error(DWORD rc) {
+  LPSTR message;
+  FormatMessage(
+    FORMAT_MESSAGE_ALLOCATE_BUFFER |
+    FORMAT_MESSAGE_FROM_SYSTEM |
+    FORMAT_MESSAGE_IGNORE_INSERTS,
+    NULL,
+    rc,
+    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
+    (LPTSTR) &message,
+    0,
+    NULL
+  );
+
+  return message;
+}
+
+static void push_win32_error(lua_State *L, DWORD rc) {
+  LPSTR message = win32_error(rc);
+  lua_pushstring(L, message);
+  LocalFree(message);
+}
+#endif
+
 static int f_poll_event(lua_State *L) {
   char buf[16];
-  int mx, my, wx, wy;
+  int mx, my, w, h;
   SDL_Event e;
+  SDL_Event event_plus;
 
 top:
   if ( !SDL_PollEvent(&e) ) {
@@ -130,8 +191,8 @@ top:
         ren_resize_window();
         lua_pushstring(L, "resized");
         /* The size below will be in points. */
-        lua_pushnumber(L, e.window.data1);
-        lua_pushnumber(L, e.window.data2);
+        lua_pushinteger(L, e.window.data1);
+        lua_pushinteger(L, e.window.data2);
         return 3;
       } else if (e.window.event == SDL_WINDOWEVENT_EXPOSED) {
         rencache_invalidate();
@@ -146,6 +207,9 @@ top:
       } else if (e.window.event == SDL_WINDOWEVENT_RESTORED) {
         lua_pushstring(L, "restored");
         return 1;
+      } else if (e.window.event == SDL_WINDOWEVENT_LEAVE) {
+        lua_pushstring(L, "mouseleft");
+        return 1;
       }
       if (e.window.event == SDL_WINDOWEVENT_FOCUS_LOST) {
         lua_pushstring(L, "focuslost");
@@ -160,12 +224,11 @@ top:
       goto top;
 
     case SDL_DROPFILE:
-      SDL_GetGlobalMouseState(&mx, &my);
-      SDL_GetWindowPosition(window, &wx, &wy);
+      SDL_GetMouseState(&mx, &my);
       lua_pushstring(L, "filedropped");
       lua_pushstring(L, e.drop.file);
-      lua_pushnumber(L, mx - wx);
-      lua_pushnumber(L, my - wy);
+      lua_pushinteger(L, mx);
+      lua_pushinteger(L, my);
       SDL_free(e.drop.file);
       return 4;
 
@@ -201,26 +264,42 @@ top:
       lua_pushstring(L, e.text.text);
       return 2;
 
+    case SDL_TEXTEDITING:
+      lua_pushstring(L, "textediting");
+      lua_pushstring(L, e.edit.text);
+      lua_pushinteger(L, e.edit.start);
+      lua_pushinteger(L, e.edit.length);
+      return 4;
+
+#if SDL_VERSION_ATLEAST(2, 0, 22)
+    case SDL_TEXTEDITING_EXT:
+      lua_pushstring(L, "textediting");
+      lua_pushstring(L, e.editExt.text);
+      lua_pushinteger(L, e.editExt.start);
+      lua_pushinteger(L, e.editExt.length);
+      SDL_free(e.editExt.text);
+      return 4;
+#endif
+
     case SDL_MOUSEBUTTONDOWN:
       if (e.button.button == 1) { SDL_CaptureMouse(1); }
       lua_pushstring(L, "mousepressed");
       lua_pushstring(L, button_name(e.button.button));
-      lua_pushnumber(L, e.button.x);
-      lua_pushnumber(L, e.button.y);
-      lua_pushnumber(L, e.button.clicks);
+      lua_pushinteger(L, e.button.x);
+      lua_pushinteger(L, e.button.y);
+      lua_pushinteger(L, e.button.clicks);
       return 5;
 
     case SDL_MOUSEBUTTONUP:
       if (e.button.button == 1) { SDL_CaptureMouse(0); }
       lua_pushstring(L, "mousereleased");
       lua_pushstring(L, button_name(e.button.button));
-      lua_pushnumber(L, e.button.x);
-      lua_pushnumber(L, e.button.y);
+      lua_pushinteger(L, e.button.x);
+      lua_pushinteger(L, e.button.y);
       return 4;
 
     case SDL_MOUSEMOTION:
       SDL_PumpEvents();
-      SDL_Event event_plus;
       while (SDL_PeepEvents(&event_plus, 1, SDL_GETEVENT, SDL_MOUSEMOTION, SDL_MOUSEMOTION) > 0) {
         e.motion.x = event_plus.motion.x;
         e.motion.y = event_plus.motion.y;
@@ -228,37 +307,60 @@ top:
         e.motion.yrel += event_plus.motion.yrel;
       }
       lua_pushstring(L, "mousemoved");
-      lua_pushnumber(L, e.motion.x);
-      lua_pushnumber(L, e.motion.y);
-      lua_pushnumber(L, e.motion.xrel);
-      lua_pushnumber(L, e.motion.yrel);
+      lua_pushinteger(L, e.motion.x);
+      lua_pushinteger(L, e.motion.y);
+      lua_pushinteger(L, e.motion.xrel);
+      lua_pushinteger(L, e.motion.yrel);
       return 5;
 
     case SDL_MOUSEWHEEL:
       lua_pushstring(L, "mousewheel");
-      lua_pushnumber(L, e.wheel.y);
-      return 2;
+#if SDL_VERSION_ATLEAST(2, 0, 18)
+      lua_pushnumber(L, e.wheel.preciseY);
+      // Use -x to keep consistency with vertical scrolling values (e.g. shift+scroll)
+      lua_pushnumber(L, -e.wheel.preciseX);
+#else
+      lua_pushinteger(L, e.wheel.y);
+      lua_pushinteger(L, -e.wheel.x);
+#endif
+      return 3;
 
-    case SDL_USEREVENT:
-      lua_pushstring(L, "dirchange");
-      lua_pushnumber(L, e.user.code >> 16);
-      switch (e.user.code & 0xffff) {
-        case DMON_ACTION_DELETE:
-          lua_pushstring(L, "delete");
-          break;
-        case DMON_ACTION_CREATE:
-          lua_pushstring(L, "create");
-          break;
-        case DMON_ACTION_MODIFY:
-          lua_pushstring(L, "modify");
-          break;
-        default:
-          return luaL_error(L, "unknown dmon event action: %d", e.user.code & 0xffff);
-      }
-      lua_pushstring(L, e.user.data1);
-      free(e.user.data1);
+      case SDL_FINGERDOWN:
+      SDL_GetWindowSize(window, &w, &h);
+
+      lua_pushstring(L, "touchpressed");
+      lua_pushinteger(L, (lua_Integer)(e.tfinger.x * w));
+      lua_pushinteger(L, (lua_Integer)(e.tfinger.y * h));
+      lua_pushinteger(L, e.tfinger.fingerId);
       return 4;
 
+    case SDL_FINGERUP:
+      SDL_GetWindowSize(window, &w, &h);
+
+      lua_pushstring(L, "touchreleased");
+      lua_pushinteger(L, (lua_Integer)(e.tfinger.x * w));
+      lua_pushinteger(L, (lua_Integer)(e.tfinger.y * h));
+      lua_pushinteger(L, e.tfinger.fingerId);
+      return 4;
+
+    case SDL_FINGERMOTION:
+      SDL_PumpEvents();
+      while (SDL_PeepEvents(&event_plus, 1, SDL_GETEVENT, SDL_FINGERMOTION, SDL_FINGERMOTION) > 0) {
+        e.tfinger.x = event_plus.tfinger.x;
+        e.tfinger.y = event_plus.tfinger.y;
+        e.tfinger.dx += event_plus.tfinger.dx;
+        e.tfinger.dy += event_plus.tfinger.dy;
+      }
+      SDL_GetWindowSize(window, &w, &h);
+
+      lua_pushstring(L, "touchmoved");
+      lua_pushinteger(L, (lua_Integer)(e.tfinger.x * w));
+      lua_pushinteger(L, (lua_Integer)(e.tfinger.y * h));
+      lua_pushinteger(L, (lua_Integer)(e.tfinger.dx * w));
+      lua_pushinteger(L, (lua_Integer)(e.tfinger.dy * h));
+      lua_pushinteger(L, e.tfinger.fingerId);
+      return 6;
+
     default:
       goto top;
   }
@@ -356,10 +458,10 @@ static int f_get_window_size(lua_State *L) {
   int x, y, w, h;
   SDL_GetWindowSize(window, &w, &h);
   SDL_GetWindowPosition(window, &x, &y);
-  lua_pushnumber(L, w);
-  lua_pushnumber(L, h);
-  lua_pushnumber(L, x);
-  lua_pushnumber(L, y);
+  lua_pushinteger(L, w);
+  lua_pushinteger(L, h);
+  lua_pushinteger(L, x);
+  lua_pushinteger(L, y);
   return 4;
 }
 
@@ -397,6 +499,36 @@ static int f_get_window_mode(lua_State *L) {
   return 1;
 }
 
+static int f_set_text_input_rect(lua_State *L) {
+  SDL_Rect rect;
+  rect.x = luaL_checknumber(L, 1);
+  rect.y = luaL_checknumber(L, 2);
+  rect.w = luaL_checknumber(L, 3);
+  rect.h = luaL_checknumber(L, 4);
+  SDL_SetTextInputRect(&rect);
+  return 0;
+}
+
+static int f_clear_ime(lua_State *L) {
+#if SDL_VERSION_ATLEAST(2, 0, 22)
+  SDL_ClearComposition();
+#endif
+  return 0;
+}
+
+
+static int f_raise_window(lua_State *L) {
+  /*
+    SDL_RaiseWindow should be enough but on some window managers like the
+    one used on Gnome the window needs to first have input focus in order
+    to allow the window to be focused. Also on wayland the raise window event
+    may not always be obeyed.
+  */
+  SDL_SetWindowInputFocus(window);
+  SDL_RaiseWindow(window);
+  return 0;
+}
+
 
 static int f_show_fatal_error(lua_State *L) {
   const char *title = luaL_checkstring(L, 1);
@@ -404,19 +536,8 @@ static int f_show_fatal_error(lua_State *L) {
 
 #ifdef _WIN32
   MessageBox(0, msg, title, MB_OK | MB_ICONERROR);
-
 #else
-  SDL_MessageBoxButtonData buttons[] = {
-    { SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT, 0, "Ok" },
-  };
-  SDL_MessageBoxData data = {
-    .title = title,
-    .message = msg,
-    .numbuttons = 1,
-    .buttons = buttons,
-  };
-  int buttonid;
-  SDL_ShowMessageBox(&data, &buttonid);
+  SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR, title, msg, NULL);
 #endif
   return 0;
 }
@@ -427,29 +548,14 @@ static int f_rmdir(lua_State *L) {
   const char *path = luaL_checkstring(L, 1);
 
 #ifdef _WIN32
-  int deleted = RemoveDirectoryA(path);
-  if(deleted > 0) {
+  LPWSTR wpath = utfconv_utf8towc(path);
+  int deleted = RemoveDirectoryW(wpath);
+  free(wpath);
+  if (deleted > 0) {
     lua_pushboolean(L, 1);
   } else {
-    DWORD error_code = GetLastError();
-    LPVOID message;
-
-    FormatMessage(
-      FORMAT_MESSAGE_ALLOCATE_BUFFER |
-      FORMAT_MESSAGE_FROM_SYSTEM |
-      FORMAT_MESSAGE_IGNORE_INSERTS,
-      NULL,
-      error_code,
-      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-      (LPTSTR) &message,
-      0,
-      NULL
-    );
-
     lua_pushboolean(L, 0);
-    lua_pushlstring(L, (LPCTSTR)message, lstrlen((LPCTSTR)message));
-    LocalFree(message);
-
+    push_win32_error(L, GetLastError());
     return 2;
   }
 #else
@@ -470,8 +576,15 @@ static int f_rmdir(lua_State *L) {
 
 static int f_chdir(lua_State *L) {
   const char *path = luaL_checkstring(L, 1);
+#ifdef _WIN32
+  LPWSTR wpath = utfconv_utf8towc(path);
+  if (wpath == NULL) { return luaL_error(L, UTFCONV_ERROR_INVALID_CONVERSION ); }
+  int err = _wchdir(wpath);
+  free(wpath);
+#else
   int err = chdir(path);
-  if (err) { luaL_error(L, "chdir() failed"); }
+#endif
+  if (err) { luaL_error(L, "chdir() failed: %s", strerror(errno)); }
   return 0;
 }
 
@@ -479,6 +592,57 @@ static int f_chdir(lua_State *L) {
 static int f_list_dir(lua_State *L) {
   const char *path = luaL_checkstring(L, 1);
 
+#ifdef _WIN32
+  lua_settop(L, 1);
+  if (path[0] == 0 || strchr("\\/", path[strlen(path) - 1]) != NULL)
+    lua_pushstring(L, "*");
+  else
+    lua_pushstring(L, "/*");
+
+  lua_concat(L, 2);
+  path = lua_tostring(L, -1);
+
+  LPWSTR wpath = utfconv_utf8towc(path);
+  if (wpath == NULL) {
+    lua_pushnil(L);
+    lua_pushstring(L, UTFCONV_ERROR_INVALID_CONVERSION);
+    return 2;
+  }
+
+  WIN32_FIND_DATAW fd;
+  HANDLE find_handle = FindFirstFileExW(wpath, FindExInfoBasic, &fd, FindExSearchNameMatch, NULL, 0);
+  free(wpath);
+  if (find_handle == INVALID_HANDLE_VALUE) {
+    lua_pushnil(L);
+    push_win32_error(L, GetLastError());
+    return 2;
+  }
+
+  char mbpath[MAX_PATH * 4]; // utf-8 spans 4 bytes at most
+  int len, i = 1;
+  lua_newtable(L);
+
+  do
+  {
+    if (wcscmp(fd.cFileName, L".") == 0) { continue; }
+    if (wcscmp(fd.cFileName, L"..") == 0) { continue; }
+
+    len = WideCharToMultiByte(CP_UTF8, 0, fd.cFileName, -1, mbpath, MAX_PATH * 4, NULL, NULL);
+    if (len == 0) { break; }
+    lua_pushlstring(L, mbpath, len - 1); // len includes \0
+    lua_rawseti(L, -2, i++);
+  } while (FindNextFileW(find_handle, &fd));
+
+  if (GetLastError() != ERROR_NO_MORE_FILES) {
+    lua_pushnil(L);
+    push_win32_error(L, GetLastError());
+    FindClose(find_handle);
+    return 2;
+  }
+
+  FindClose(find_handle);
+  return 1;
+#else
   DIR *dir = opendir(path);
   if (!dir) {
     lua_pushnil(L);
@@ -499,17 +663,29 @@ static int f_list_dir(lua_State *L) {
 
   closedir(dir);
   return 1;
+#endif
 }
 
 
 #ifdef _WIN32
-  #include <windows.h>
-  #define realpath(x, y) _fullpath(y, x, MAX_PATH)
+  #define realpath(x, y) _wfullpath(y, x, MAX_PATH)
 #endif
 
 static int f_absolute_path(lua_State *L) {
   const char *path = luaL_checkstring(L, 1);
+#ifdef _WIN32
+  LPWSTR wpath = utfconv_utf8towc(path);
+  if (!wpath) { return 0; }
+
+  LPWSTR wfullpath = realpath(wpath, NULL);
+  free(wpath);
+  if (!wfullpath) { return 0; }
+
+  char *res = utfconv_wctoutf8(wfullpath);
+  free(wfullpath);
+#else
   char *res = realpath(path, NULL);
+#endif
   if (!res) { return 0; }
   lua_pushstring(L, res);
   free(res);
@@ -520,8 +696,20 @@ static int f_absolute_path(lua_State *L) {
 static int f_get_file_info(lua_State *L) {
   const char *path = luaL_checkstring(L, 1);
 
+#ifdef _WIN32
+  struct _stat s;
+  LPWSTR wpath = utfconv_utf8towc(path);
+  if (wpath == NULL) {
+    lua_pushnil(L);
+    lua_pushstring(L, UTFCONV_ERROR_INVALID_CONVERSION);
+    return 2;
+  }
+  int err = _wstat(wpath, &s);
+  free(wpath);
+#else
   struct stat s;
   int err = stat(path, &s);
+#endif
   if (err < 0) {
     lua_pushnil(L);
     lua_pushstring(L, strerror(errno));
@@ -529,10 +717,10 @@ static int f_get_file_info(lua_State *L) {
   }
 
   lua_newtable(L);
-  lua_pushnumber(L, s.st_mtime);
+  lua_pushinteger(L, s.st_mtime);
   lua_setfield(L, -2, "modified");
 
-  lua_pushnumber(L, s.st_size);
+  lua_pushinteger(L, s.st_size);
   lua_setfield(L, -2, "size");
 
   if (S_ISREG(s.st_mode)) {
@@ -544,6 +732,14 @@ static int f_get_file_info(lua_State *L) {
   }
   lua_setfield(L, -2, "type");
 
+#if __linux__
+  if (S_ISDIR(s.st_mode)) {
+    if (lstat(path, &s) == 0) {
+      lua_pushboolean(L, S_ISLNK(s.st_mode));
+      lua_setfield(L, -2, "symlink");
+    }
+  }
+#endif
   return 1;
 }
 
@@ -568,30 +764,41 @@ static struct f_type_names fs_names[] = {
   { 0x0,        NULL        },
 };
 
+#endif
+
 static int f_get_fs_type(lua_State *L) {
-  const char *path = luaL_checkstring(L, 1);
-  struct statfs buf;
-  int status = statfs(path, &buf);
-  if (status != 0) {
-    return luaL_error(L, "error calling statfs on %s", path);
-  }
-  for (int i = 0; fs_names[i].magic; i++) {
-    if (fs_names[i].magic == buf.f_type) {
-      lua_pushstring(L, fs_names[i].name);
-      return 1;
+  #if __linux__
+    const char *path = luaL_checkstring(L, 1);
+    struct statfs buf;
+    int status = statfs(path, &buf);
+    if (status != 0) {
+      return luaL_error(L, "error calling statfs on %s", path);
     }
-  }
+    for (int i = 0; fs_names[i].magic; i++) {
+      if (fs_names[i].magic == buf.f_type) {
+        lua_pushstring(L, fs_names[i].name);
+        return 1;
+      }
+    }
+  #endif
   lua_pushstring(L, "unknown");
   return 1;
 }
-#endif
 
 
 static int f_mkdir(lua_State *L) {
   const char *path = luaL_checkstring(L, 1);
 
 #ifdef _WIN32
-  int err = _mkdir(path);
+  LPWSTR wpath = utfconv_utf8towc(path);
+  if (wpath == NULL) {
+    lua_pushboolean(L, 0);
+    lua_pushstring(L, UTFCONV_ERROR_INVALID_CONVERSION);
+    return 2;
+  }
+
+  int err = _wmkdir(wpath);
+  free(wpath);
 #else
   int err = mkdir(path, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
 #endif
@@ -622,6 +829,16 @@ static int f_set_clipboard(lua_State *L) {
 }
 
 
+static int f_get_process_id(lua_State *L) {
+#ifdef _WIN32
+  lua_pushinteger(L, GetCurrentProcessId());
+#else
+  lua_pushinteger(L, getpid());
+#endif
+  return 1;
+}
+
+
 static int f_get_time(lua_State *L) {
   double n = SDL_GetPerformanceCounter() / (double) SDL_GetPerformanceFrequency();
   lua_pushnumber(L, n);
@@ -653,56 +870,32 @@ static int f_exec(lua_State *L) {
   return 0;
 }
 
-
 static int f_fuzzy_match(lua_State *L) {
   size_t strLen, ptnLen;
   const char *str = luaL_checklstring(L, 1, &strLen);
   const char *ptn = luaL_checklstring(L, 2, &ptnLen);
-  bool files = false;
-  if (lua_gettop(L) > 2 && lua_isboolean(L,3))
-    files = lua_toboolean(L, 3);
-
-  int score = 0;
-  int run = 0;
-
-  // Match things *backwards*. This allows for better matching on filenames than the above
+  // If true match things *backwards*. This allows for better matching on filenames than the above
   // function. For example, in the lite project, opening "renderer" has lib/font_render/build.sh
   // as the first result, rather than src/renderer.c. Clearly that's wrong.
-  if (files) {
-    const char* strEnd = str + strLen - 1;
-    const char* ptnEnd = ptn + ptnLen - 1;
-    while (strEnd >= str && ptnEnd >= ptn) {
-      while (*strEnd == ' ') { strEnd--; }
-      while (*ptnEnd == ' ') { ptnEnd--; }
-      if (tolower(*strEnd) == tolower(*ptnEnd)) {
-        score += run * 10 - (*strEnd != *ptnEnd);
-        run++;
-        ptnEnd--;
-      } else {
-        score -= 10;
-        run = 0;
-      }
-      strEnd--;
+  bool files = lua_gettop(L) > 2 && lua_isboolean(L,3) && lua_toboolean(L, 3);
+  int score = 0, run = 0, increment = files ? -1 : 1;
+  const char* strTarget = files ? str + strLen - 1 : str;
+  const char* ptnTarget = files ? ptn + ptnLen - 1 : ptn;
+  while (strTarget >= str && ptnTarget >= ptn && *strTarget && *ptnTarget) {
+    while (strTarget >= str && *strTarget == ' ') { strTarget += increment; }
+    while (ptnTarget >= ptn && *ptnTarget == ' ') { ptnTarget += increment; }
+    if (tolower(*strTarget) == tolower(*ptnTarget)) {
+      score += run * 10 - (*strTarget != *ptnTarget);
+      run++;
+      ptnTarget += increment;
+    } else {
+      score -= 10;
+      run = 0;
     }
-    if (ptnEnd >= ptn) { return 0; }
-  } else {
-    while (*str && *ptn) {
-      while (*str == ' ') { str++; }
-      while (*ptn == ' ') { ptn++; }
-      if (tolower(*str) == tolower(*ptn)) {
-        score += run * 10 - (*str != *ptn);
-        run++;
-        ptn++;
-      } else {
-        score -= 10;
-        run = 0;
-      }
-      str++;
-    }
-    if (*ptn) { return 0; }
+    strTarget += increment;
   }
-
-  lua_pushnumber(L, score - (int)strLen);
+  if (ptnTarget >= ptn && *ptnTarget) { return 0; }
+  lua_pushinteger(L, score - (int)strLen * 10);
   return 1;
 }
 
@@ -713,33 +906,112 @@ static int f_set_window_opacity(lua_State *L) {
   return 1;
 }
 
-static int f_watch_dir(lua_State *L) {
-  const char *path = luaL_checkstring(L, 1);
-  const int recursive = lua_toboolean(L, 2);
-  uint32_t dmon_flags = (recursive ? DMON_WATCHFLAGS_RECURSIVE : 0);
-  dmon_watch_id watch_id = dmon_watch(path, dirmonitor_watch_callback, dmon_flags, NULL);
-  if (watch_id.id == 0) { luaL_error(L, "directory monitoring watch failed"); }
-  lua_pushnumber(L, watch_id.id);
-  return 1;
+typedef void (*fptr)(void);
+
+typedef struct lua_function_node {
+  const char *symbol;
+  fptr address;
+} lua_function_node;
+
+#define P(FUNC) { "lua_" #FUNC, (fptr)(lua_##FUNC) }
+#define U(FUNC) { "luaL_" #FUNC, (fptr)(luaL_##FUNC) }
+static void* api_require(const char* symbol) {
+  static lua_function_node nodes[] = {
+    P(atpanic), P(checkstack),
+    P(close), P(concat), P(copy), P(createtable), P(dump),
+    P(error),  P(gc), P(getallocf),  P(getfield),
+    P(gethook), P(gethookcount), P(gethookmask), P(getinfo), P(getlocal),
+    P(getmetatable), P(getstack), P(gettable), P(gettop), P(getupvalue),
+    P(isnumber), P(isstring), P(isuserdata),
+    P(load), P(newstate), P(newthread), P(next),
+    P(pushboolean), P(pushcclosure), P(pushfstring), P(pushinteger),
+    P(pushlightuserdata), P(pushlstring), P(pushnil), P(pushnumber),
+    P(pushstring), P(pushthread),  P(pushvalue),
+    P(pushvfstring), P(rawequal), P(rawget), P(rawgeti),
+    P(rawset), P(rawseti), P(resume),
+    P(setallocf), P(setfield), P(sethook), P(setlocal),
+    P(setmetatable), P(settable), P(settop), P(setupvalue),
+    P(status), P(tocfunction), P(tointegerx), P(tolstring), P(toboolean),
+    P(tonumberx), P(topointer), P(tothread),  P(touserdata),
+    P(type), P(typename), P(upvalueid), P(upvaluejoin), P(version), P(xmove),
+    U(getmetafield), U(callmeta), U(argerror), U(checknumber), U(optnumber),
+    U(checkinteger), U(checkstack), U(checktype), U(checkany),
+    U(newmetatable), U(setmetatable), U(testudata), U(checkudata), U(where),
+    U(error), U(fileresult), U(execresult), U(ref), U(unref), U(loadstring),
+    U(newstate), U(setfuncs), U(buffinit), U(addlstring), U(addstring),
+    U(addvalue), U(pushresult), U(openlibs), {"api_load_libs", (void*)(api_load_libs)},
+    #if LUA_VERSION_NUM >= 502
+    P(absindex), P(arith), P(callk), P(compare), P(getglobal),
+    P(len), P(pcallk), P(rawgetp), P(rawlen), P(rawsetp), P(setglobal),
+    P(iscfunction), P(yieldk),
+    U(checkversion_), U(tolstring), U(len), U(getsubtable), U(prepbuffsize),
+    U(pushresultsize), U(buffinitsize), U(checklstring), U(checkoption), U(gsub), U(loadbufferx),
+    U(loadfilex), U(optinteger), U(optlstring), U(requiref), U(traceback),
+    #else
+    P(objlen),
+    #endif
+    #if LUA_VERSION_NUM >= 504
+    P(newuserdatauv), P(setiuservalue), P(getiuservalue)
+    #else
+    P(newuserdata), P(setuservalue), P(getuservalue)
+    #endif
+
+  };
+  for (size_t i = 0; i < sizeof(nodes) / sizeof(lua_function_node); ++i) {
+    if (strcmp(nodes[i].symbol, symbol) == 0)
+      return *(void**)(&nodes[i].address);
+  }
+  return NULL;
 }
 
-#if __linux__
-static int f_watch_dir_add(lua_State *L) {
-  dmon_watch_id watch_id;
-  watch_id.id = luaL_checkinteger(L, 1);
-  const char *subdir = luaL_checkstring(L, 2);
-  lua_pushboolean(L, dmon_watch_add(watch_id, subdir));
-  return 1;
+static int f_library_gc(lua_State *L) {
+  lua_getfield(L, 1, "handle");
+  void* handle = lua_touserdata(L, -1);
+  SDL_UnloadObject(handle);
+  
+  return 0;
 }
 
-static int f_watch_dir_rm(lua_State *L) {
-  dmon_watch_id watch_id;
-  watch_id.id = luaL_checkinteger(L, 1);
-  const char *subdir = luaL_checkstring(L, 2);
-  lua_pushboolean(L, dmon_watch_rm(watch_id, subdir));
-  return 1;
+static int f_load_native_plugin(lua_State *L) {
+  char entrypoint_name[512]; entrypoint_name[sizeof(entrypoint_name) - 1] = '\0';
+  int result;
+
+  const char *name = luaL_checkstring(L, 1);
+  const char *path = luaL_checkstring(L, 2);
+  void *library = SDL_LoadObject(path);
+  if (!library)
+    return (lua_pushstring(L, SDL_GetError()), lua_error(L));
+
+  lua_getglobal(L, "package");
+  lua_getfield(L, -1, "native_plugins");
+  lua_newtable(L);
+  lua_pushlightuserdata(L, library);
+  lua_setfield(L, -2, "handle");
+  luaL_setmetatable(L, API_TYPE_NATIVE_PLUGIN);
+  lua_setfield(L, -2, name);
+  lua_pop(L, 2);
+
+  const char *basename = strrchr(name, '.');
+  basename = !basename ? name : basename + 1;
+  snprintf(entrypoint_name, sizeof(entrypoint_name), "luaopen_lite_xl_%s", basename);
+  int (*ext_entrypoint) (lua_State *L, void* (*)(const char*));
+  *(void**)(&ext_entrypoint) = SDL_LoadFunction(library, entrypoint_name);
+  if (!ext_entrypoint) {
+    snprintf(entrypoint_name, sizeof(entrypoint_name), "luaopen_%s", basename);
+    int (*entrypoint)(lua_State *L);
+    *(void**)(&entrypoint) = SDL_LoadFunction(library, entrypoint_name);
+    if (!entrypoint)
+      return luaL_error(L, "Unable to load %s: Can't find %s(lua_State *L, void *XL)", name, entrypoint_name);
+    result = entrypoint(L);
+  } else {
+    result = ext_entrypoint(L, api_require);
+  }
+
+  if (!result)
+    return luaL_error(L, "Unable to load %s: entrypoint must return a value", name);
+
+  return result;
 }
-#endif
 
 #ifdef _WIN32
 #define PATHSEP '\\'
@@ -810,7 +1082,10 @@ static const luaL_Reg lib[] = {
   { "set_window_hit_test", f_set_window_hit_test },
   { "get_window_size",     f_get_window_size     },
   { "set_window_size",     f_set_window_size     },
+  { "set_text_input_rect", f_set_text_input_rect },
+  { "clear_ime",           f_clear_ime           },
   { "window_has_focus",    f_window_has_focus    },
+  { "raise_window",        f_raise_window        },
   { "show_fatal_error",    f_show_fatal_error    },
   { "rmdir",               f_rmdir               },
   { "chdir",               f_chdir               },
@@ -820,23 +1095,23 @@ static const luaL_Reg lib[] = {
   { "get_file_info",       f_get_file_info       },
   { "get_clipboard",       f_get_clipboard       },
   { "set_clipboard",       f_set_clipboard       },
+  { "get_process_id",      f_get_process_id      },
   { "get_time",            f_get_time            },
   { "sleep",               f_sleep               },
   { "exec",                f_exec                },
   { "fuzzy_match",         f_fuzzy_match         },
   { "set_window_opacity",  f_set_window_opacity  },
-  { "watch_dir",           f_watch_dir           },
+  { "load_native_plugin",  f_load_native_plugin  },
   { "path_compare",        f_path_compare        },
-#if __linux__
-  { "watch_dir_add",       f_watch_dir_add       },
-  { "watch_dir_rm",        f_watch_dir_rm        },
   { "get_fs_type",         f_get_fs_type         },
-#endif
   { NULL, NULL }
 };
 
 
 int luaopen_system(lua_State *L) {
+  luaL_newmetatable(L, API_TYPE_NATIVE_PLUGIN); 
+  lua_pushcfunction(L, f_library_gc);
+  lua_setfield(L, -2, "__gc");
   luaL_newlib(L, lib);
   return 1;
 }
