@@ -5,17 +5,18 @@
 #include "rencache.h"
 #include "renderer.h"
 
+#include <signal.h>
+
 #ifdef _WIN32
   #include <windows.h>
-#elif __linux__
+#elif defined(__linux__)
   #include <unistd.h>
-  #include <signal.h>
-#elif __APPLE__
+#elif defined(__APPLE__)
   #include <mach-o/dyld.h>
+#elif defined(__FreeBSD__)
+  #include <sys/sysctl.h>
 #endif
 
-#include "dirmonitor.h"
-
 
 SDL_Window *window;
 
@@ -34,8 +35,7 @@ static void get_exe_filename(char *buf, int sz) {
   int len = GetModuleFileName(NULL, buf, sz - 1);
   buf[len] = '\0';
 #elif __linux__
-  char path[512];
-  sprintf(path, "/proc/%d/exe", getpid());
+  char path[] = "/proc/self/exe";
   int len = readlink(path, buf, sz - 1);
   buf[len] = '\0';
 #elif __APPLE__
@@ -46,8 +46,12 @@ static void get_exe_filename(char *buf, int sz) {
   char exepath[size];
   _NSGetExecutablePath(exepath, &size);
   realpath(exepath, buf);
+#elif __FreeBSD__
+  size_t len = sz;
+  const int mib[4] = { CTL_KERN, KERN_PROC, KERN_PROC_PATHNAME, -1 };
+  sysctl(mib, 4, buf, &len, NULL, 0);
 #else
-  strcpy(buf, "./lite");
+  *buf = 0;
 #endif
 }
 
@@ -85,6 +89,35 @@ void set_macos_bundle_resources(lua_State *L);
 #endif
 #endif
 
+#ifndef LITE_ARCH_TUPLE
+  // https://learn.microsoft.com/en-us/cpp/preprocessor/predefined-macros?view=msvc-140
+  #if defined(__x86_64__) || defined(_M_AMD64) || defined(__MINGW64__)
+    #define ARCH_PROCESSOR "x86_64"
+  #elif defined(__i386__) || defined(_M_IX86) || defined(__MINGW32__)
+    #define ARCH_PROCESSOR "x86"
+  #elif defined(__aarch64__) || defined(_M_ARM64) || defined (_M_ARM64EC)
+    #define ARCH_PROCESSOR "aarch64"
+  #elif defined(__arm__) || defined(_M_ARM)
+    #define ARCH_PROCESSOR "arm"
+  #endif
+
+  #if _WIN32
+    #define ARCH_PLATFORM "windows"
+  #elif __linux__
+    #define ARCH_PLATFORM "linux"
+  #elif __FreeBSD__
+    #define ARCH_PLATFORM "freebsd"
+  #elif __APPLE__
+    #define ARCH_PLATFORM "darwin"
+  #endif
+
+  #if !defined(ARCH_PROCESSOR) || !defined(ARCH_PLATFORM)
+    #error "Please define -DLITE_ARCH_TUPLE."
+  #endif
+
+  #define LITE_ARCH_TUPLE ARCH_PROCESSOR "-" ARCH_PLATFORM
+#endif
+
 int main(int argc, char **argv) {
 #ifdef _WIN32
   HINSTANCE lib = LoadLibrary("user32.dll");
@@ -94,7 +127,10 @@ int main(int argc, char **argv) {
   signal(SIGPIPE, SIG_IGN);
 #endif
 
-  SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTS);
+  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTS) != 0) {
+    fprintf(stderr, "Error initializing sdl: %s", SDL_GetError());
+    exit(1);
+  }
   SDL_EnableScreenSaver();
   SDL_EventState(SDL_DROPFILE, SDL_ENABLE);
   atexit(SDL_Quit);
@@ -105,16 +141,39 @@ int main(int argc, char **argv) {
 #if SDL_VERSION_ATLEAST(2, 0, 5)
   SDL_SetHint(SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH, "1");
 #endif
+#if SDL_VERSION_ATLEAST(2, 0, 18)
+  SDL_SetHint(SDL_HINT_IME_SHOW_UI, "1");
+#endif
+#if SDL_VERSION_ATLEAST(2, 0, 22)
+  SDL_SetHint(SDL_HINT_IME_SUPPORT_EXTENDED_TEXT, "1");
+#endif
+
+#if SDL_VERSION_ATLEAST(2, 0, 8)
+  /* This hint tells SDL to respect borderless window as a normal window.
+  ** For example, the window will sit right on top of the taskbar instead
+  ** of obscuring it. */
+  SDL_SetHint("SDL_BORDERLESS_WINDOWED_STYLE", "1");
+#endif
+#if SDL_VERSION_ATLEAST(2, 0, 12)
+  /* This hint tells SDL to allow the user to resize a borderless windoow.
+  ** It also enables aero-snap on Windows apparently. */
+  SDL_SetHint("SDL_BORDERLESS_RESIZABLE_STYLE", "1");
+#endif
+#if SDL_VERSION_ATLEAST(2, 0, 9)
+  SDL_SetHint("SDL_MOUSE_DOUBLE_CLICK_RADIUS", "4");
+#endif
 
   SDL_DisplayMode dm;
   SDL_GetCurrentDisplayMode(0, &dm);
 
-  dirmonitor_init();
-
   window = SDL_CreateWindow(
     "", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, dm.w * 0.8, dm.h * 0.8,
     SDL_WINDOW_RESIZABLE | SDL_WINDOW_ALLOW_HIGHDPI | SDL_WINDOW_HIDDEN);
   init_window_icon();
+  if (!window) {
+    fprintf(stderr, "Error creating lite-xl window: %s", SDL_GetError());
+    exit(1);
+  }
   ren_init(window);
 
   lua_State *L;
@@ -134,12 +193,20 @@ init_lua:
   lua_pushstring(L, SDL_GetPlatform());
   lua_setglobal(L, "PLATFORM");
 
+  lua_pushstring(L, LITE_ARCH_TUPLE);
+  lua_setglobal(L, "ARCH");
+
   lua_pushnumber(L, get_scale());
   lua_setglobal(L, "SCALE");
 
   char exename[2048];
   get_exe_filename(exename, sizeof(exename));
-  lua_pushstring(L, exename);
+  if (*exename) {
+    lua_pushstring(L, exename);
+  } else {
+    // get_exe_filename failed
+    lua_pushstring(L, argv[0]);
+  }
   lua_setglobal(L, "EXEFILE");
 
 #ifdef __APPLE__
@@ -193,7 +260,6 @@ init_lua:
 
   lua_close(L);
   ren_free_window_resources();
-  dirmonitor_deinit();
 
   return EXIT_SUCCESS;
 }
