@@ -1,6 +1,19 @@
 #include <stddef.h>
 #include <stdint.h>
 #include <stdio.h>
+#include <stdbool.h>
+#include <string.h>
+
+#ifdef _MSC_VER
+  #ifndef alignof
+    #define alignof _Alignof
+  #endif
+  /* max_align_t is a compiler defined type, but
+  ** MSVC doesn't provide it, so we'll have to improvise */
+  typedef long double max_align_t;
+#else
+  #include <stdalign.h>
+#endif
 
 #include <lauxlib.h>
 #include "rencache.h"
@@ -16,32 +29,20 @@
 #define COMMAND_BUF_SIZE (1024 * 512)
 #define COMMAND_BARE_SIZE offsetof(Command, text)
 
-enum { SET_CLIP, DRAW_TEXT, DRAW_RECT, DRAW_TEXT_SUBPIXEL };
+enum { SET_CLIP, DRAW_TEXT, DRAW_RECT };
 
 typedef struct {
   int8_t type;
   int8_t tab_size;
-  int8_t subpixel_scale;
-  int8_t x_subpixel_offset;
   int32_t size;
   RenRect rect;
   RenColor color;
-  FontDesc *font_desc;
-  CPReplaceTable *replacements;
-  RenColor replace_color;
-  char text[0];
+  RenFont *fonts[FONT_FALLBACK_MAX];
+  float text_x;
+  size_t len;
+  char text[];
 } Command;
 
-#define FONT_REFS_MAX 12
-struct FontRef {
-  FontDesc *font_desc;
-  int index;
-};
-typedef struct FontRef FontRef;
-FontRef font_refs[FONT_REFS_MAX];
-int font_refs_len = 0;
-
-
 static unsigned cells_buf1[CELLS_X * CELLS_Y];
 static unsigned cells_buf2[CELLS_X * CELLS_Y];
 static unsigned *cells_prev = cells_buf1;
@@ -52,36 +53,9 @@ static int command_buf_idx;
 static RenRect screen_rect;
 static bool show_debug;
 
+static inline int rencache_min(int a, int b) { return a < b ? a : b; }
+static inline int rencache_max(int a, int b) { return a > b ? a : b; }
 
-static inline int min(int a, int b) { return a < b ? a : b; }
-static inline int max(int a, int b) { return a > b ? a : b; }
-
-static int font_refs_add(lua_State *L, FontDesc *font_desc, int index) {
-  for (int i = 0; i < font_refs_len; i++) {
-    if (font_refs[i].font_desc == font_desc) {
-      return font_refs[i].index;
-    }
-  }
-
-  if (font_refs_len >= FONT_REFS_MAX) {
-    fprintf(stderr, "Warning: (" __FILE__ "): exhausted font reference buffer\n");
-    return LUA_NOREF;
-  }
-
-  lua_pushvalue(L, index);
-  int ref = luaL_ref(L, LUA_REGISTRYINDEX);
-
-  font_refs[font_refs_len++] = (FontRef) { font_desc, ref };
-  return ref;
-}
-
-
-static void font_refs_clear(lua_State *L) {
-  for (int i = 0; i < font_refs_len; i++) {
-    luaL_unref(L, LUA_REGISTRYINDEX, font_refs[i].index);
-  }
-  font_refs_len = 0;
-}
 
 /* 32bit fnv-1a hash */
 #define HASH_INITIAL 2166136261
@@ -106,24 +80,26 @@ static inline bool rects_overlap(RenRect a, RenRect b) {
 
 
 static RenRect intersect_rects(RenRect a, RenRect b) {
-  int x1 = max(a.x, b.x);
-  int y1 = max(a.y, b.y);
-  int x2 = min(a.x + a.width, b.x + b.width);
-  int y2 = min(a.y + a.height, b.y + b.height);
-  return (RenRect) { x1, y1, max(0, x2 - x1), max(0, y2 - y1) };
+  int x1 = rencache_max(a.x, b.x);
+  int y1 = rencache_max(a.y, b.y);
+  int x2 = rencache_min(a.x + a.width, b.x + b.width);
+  int y2 = rencache_min(a.y + a.height, b.y + b.height);
+  return (RenRect) { x1, y1, rencache_max(0, x2 - x1), rencache_max(0, y2 - y1) };
 }
 
 
 static RenRect merge_rects(RenRect a, RenRect b) {
-  int x1 = min(a.x, b.x);
-  int y1 = min(a.y, b.y);
-  int x2 = max(a.x + a.width, b.x + b.width);
-  int y2 = max(a.y + a.height, b.y + b.height);
+  int x1 = rencache_min(a.x, b.x);
+  int y1 = rencache_min(a.y, b.y);
+  int x2 = rencache_max(a.x + a.width, b.x + b.width);
+  int y2 = rencache_max(a.y + a.height, b.y + b.height);
   return (RenRect) { x1, y1, x2 - x1, y2 - y1 };
 }
 
 
 static Command* push_command(int type, int size) {
+  size_t alignment = alignof(max_align_t) - 1;
+  size = (size + alignment) & ~alignment;
   Command *cmd = (Command*) (command_buf + command_buf_idx);
   int n = command_buf_idx + size;
   if (n > COMMAND_BUF_SIZE) {
@@ -131,7 +107,7 @@ static Command* push_command(int type, int size) {
     return NULL;
   }
   command_buf_idx = n;
-  memset(cmd, 0, COMMAND_BARE_SIZE);
+  memset(cmd, 0, size);
   cmd->type = type;
   cmd->size = size;
   return cmd;
@@ -170,35 +146,24 @@ void rencache_draw_rect(RenRect rect, RenColor color) {
   }
 }
 
-int rencache_draw_text(lua_State *L, FontDesc *font_desc, int font_index,
-  const char *text, int x, int y, RenColor color, bool draw_subpixel,
-  CPReplaceTable *replacements, RenColor replace_color)
+float rencache_draw_text(RenFont **fonts, const char *text, size_t len, float x, int y, RenColor color)
 {
-  int subpixel_scale;
-  int w_subpixel = ren_get_font_width(font_desc, text, &subpixel_scale);
-  RenRect rect;
-  rect.x = (draw_subpixel ? ren_font_subpixel_round(x, subpixel_scale, -1) : x);
-  rect.y = y;
-  rect.width = ren_font_subpixel_round(w_subpixel, subpixel_scale, 0);
-  rect.height = ren_get_font_height(font_desc);
-
-  if (rects_overlap(screen_rect, rect) && font_refs_add(L, font_desc, font_index) >= 0) {
-    int sz = strlen(text) + 1;
-    Command *cmd = push_command(draw_subpixel ? DRAW_TEXT_SUBPIXEL : DRAW_TEXT, COMMAND_BARE_SIZE + sz);
+  float width = ren_font_group_get_width(fonts, text, len);
+  RenRect rect = { x, y, (int)width, ren_font_group_get_height(fonts) };
+  if (rects_overlap(screen_rect, rect)) {
+    int sz = len + 1;
+    Command *cmd = push_command(DRAW_TEXT, COMMAND_BARE_SIZE + sz);
     if (cmd) {
       memcpy(cmd->text, text, sz);
       cmd->color = color;
-      cmd->font_desc = font_desc;
+      memcpy(cmd->fonts, fonts, sizeof(RenFont*)*FONT_FALLBACK_MAX);
       cmd->rect = rect;
-      cmd->subpixel_scale = (draw_subpixel ? subpixel_scale : 1);
-      cmd->x_subpixel_offset = x - subpixel_scale * rect.x;
-      cmd->tab_size = font_desc_get_tab_size(font_desc);
-      cmd->replacements = replacements;
-      cmd->replace_color = replace_color;
+      cmd->text_x = x;
+      cmd->len = len;
+      cmd->tab_size = ren_font_group_get_tab_size(fonts);
     }
   }
-
-  return x + (draw_subpixel ? w_subpixel : rect.width);
+  return x + width;
 }
 
 
@@ -207,7 +172,7 @@ void rencache_invalidate(void) {
 }
 
 
-void rencache_begin_frame(lua_State *L) {
+void rencache_begin_frame() {
   /* reset all cells if the screen width/height has changed */
   int w, h;
   ren_get_size(&w, &h);
@@ -216,7 +181,6 @@ void rencache_begin_frame(lua_State *L) {
     screen_rect.height = h;
     rencache_invalidate();
   }
-  font_refs_clear(L);
 }
 
 
@@ -249,7 +213,7 @@ static void push_rect(RenRect r, int *count) {
 }
 
 
-void rencache_end_frame(lua_State *L) {
+void rencache_end_frame() {
   /* update cells from commands */
   Command *cmd = NULL;
   RenRect cr = screen_rect;
@@ -303,15 +267,8 @@ void rencache_end_frame(lua_State *L) {
           ren_draw_rect(cmd->rect, cmd->color);
           break;
         case DRAW_TEXT:
-          font_desc_set_tab_size(cmd->font_desc, cmd->tab_size);
-          ren_draw_text(cmd->font_desc, cmd->text, cmd->rect.x, cmd->rect.y, cmd->color,
-            cmd->replacements, cmd->replace_color);
-          break;
-        case DRAW_TEXT_SUBPIXEL:
-          font_desc_set_tab_size(cmd->font_desc, cmd->tab_size);
-          ren_draw_text_subpixel(cmd->font_desc, cmd->text,
-            cmd->subpixel_scale * cmd->rect.x + cmd->x_subpixel_offset, cmd->rect.y, cmd->color,
-            cmd->replacements, cmd->replace_color);
+          ren_font_group_set_tab_size(cmd->fonts, cmd->tab_size);
+          ren_draw_text(cmd->fonts, cmd->text, cmd->len, cmd->text_x, cmd->rect.y, cmd->color);
           break;
       }
     }
